﻿// Copyright (c) 2005 Prodige Software corporation. All rights reserved.

// Portions copyright (c) 2004, Arron Washington. All rights reserved.
/*
Arron Washington copyright restrictions:

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the The Radical Edwards nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#region Using directives

using System;
using System.Collections;
using System.Text;
using System.Xml;
using System.IO;
using System.Drawing;
using Prodige.Drawing;
using Prodige.Drawing.Styles;
using Path = Prodige.Drawing.Path;
using SvgNet;
using SvgNet.SvgElements;
using SvgNet.SvgTypes;
#endregion

/*
 * Implementation notes:
 * 
 * The underlying SVG parser used by this converter is deficient. Specifically,
 * all attribute values are stored in a hashtable per node, instead of strongly typed fields.
 * While many attribute values may be inherited and should be in a table, there are also many
 * that cannot be inherited, and should be stored directly on the object.
 * 
 * A common flaw in the parser was to assume that an attribute retrieved from the table was
 * of a specific type, when there was no enforcement to ensure that the attribute was stored
 * as that type. In short, no true invariant was used for data storage. I have corrected all
 * the errors like this I could find, but a more robust architecture requires a major overhaul
 * of the entire design.
 * 
 * The second problem is simply lack of implementation. Every possible enum-like value in
 * SVG should have a corresponding enum in the parser library. There should be no ability to store
 * two different data types, such as color or string, into the same attribute slot.
 * 
 * While the SharpVectors project provides a more accurate and robust parsing of SVG files,
 * it suffers from overengineering for this type of simple converter. At some point we have to decide
 * if we wish to enhance the flawed parser used here, or switch to the SharpVectors parser.
 * 
 * Frank Hileman, 2005
 */

namespace Prodige.Svg
{
    /// <summary>
    /// Converts SVG files and streams to VG.net Pictures.
    /// </summary>
    public class SvgImporter : PictureImporter
    {
		/// <summary>The Styles to be added to the new Picture.</summary>
		private ArrayList styles;

        //JPV : CSS Class list
        private ArrayList stylesClass;

 		public SvgImporter()
		{
			styles = new ArrayList();
		}

		public override string FormatName
		{
			get { return "svg"; }
		}

		public override bool ExtensionSupported(string fileExtension)
		{
			// fileExtension will always be lower case
			return fileExtension == "svg";
		}

		public override Picture ConvertToPicture(string fileName)
		{
			using (Stream stream = File.Open(fileName, FileMode.Open, FileAccess.Read))
				return ConvertToPicture(stream);
		}

		/// <summary>
        /// Convert SVG to a Picture using data from a Stream.
        /// </summary>
        /// <param name="stream">A stream containing svg data.</param>
 		public Picture ConvertToPicture(Stream stream)
        {
            XmlDocument document = new XmlDocument();
			document.XmlResolver = null;
            document.Load(stream);
            return ConvertRoot(SvgFactory.LoadFromXML(document, document.DocumentElement));
        }

		private Picture ConvertRoot(SvgElement root)
		{
			Check.Operation(styles.Count == 0);
			Picture picture = new Picture();
			ConvertChildren(root, picture);
			foreach (Style style in styles)
				picture.Styles.Add(style);
			styles.Clear();
            
			// NOTE: are Width and Height on root good for anything?
//			// this means it is an "svg" element, yes the name is awful
//			SvgSvgElement svgRoot = root as SvgSvgElement;
//			if (svgRoot != null)
//			{
//				picture.Width = svgRoot.Width.Value;
//				picture.Height = svgRoot.Height.Value;
//			}
			return picture;
		}
         
        private Group ConvertGroup(SvgGroupElement svgGroup)
        {
            Group group = new Group();
			ConvertChildren(svgGroup, group);
            ConvertStrokeFill(svgGroup, group);
            return group;
        }

		private void ConvertChildren(SvgElement svg, Composite composite)
		{
			foreach (SvgElement child in svg.Children)
			{
				Element converted = ConvertChild(child);
				if (converted != null)
					composite.Elements.Add(converted);
			}
		}

        /// <summary>
        /// JPV : Initialize the style class list
        /// </summary>
        /// <param name="styleList">List of CSS styles define with CDATA</param>
        private void InitClassStyle(SvgElement styleList)
        {
            stylesClass = new ArrayList();
            
            for (int i=0; i<styleList.Children.Count; i++)
            {
                SvgClassStyleElement styleClass = (SvgClassStyleElement)styleList.Children[i];
                stylesClass.Add(styleClass);
            }
        }
        
        /// <summary>
        /// JPV : Initialize Style property with a define CSS class of the SvgStyledTransformedElement
        /// </summary>
        /// <param name="pelement">Element to apply style</param>
        /// <returns>Element with style applied if possible</returns>
        private SvgStyledTransformedElement InitStyleElement(SvgElement pelement)
        {
            SvgStyledTransformedElement elementStyle = pelement as SvgStyledTransformedElement;
            string clname = pelement.Attributes["class"] as string;
            if (clname != null)
            {
                clname = "." + clname;
				foreach (SvgClassStyleElement clStyle in stylesClass)
				{
                    if (clStyle.ClassName == clname)
                    {
                        elementStyle.Style = clStyle.Style;
						break;
                    }
				}
             }
            return elementStyle;
        }
        
		private Element ConvertChild(SvgElement child)
		{
			Element converted = null;

			//JPV : Apply the style
			SvgStyledTransformedElement elementStyle = InitStyleElement(child);
			
			switch (child.Name)
			{
                //JPV : Initialize the ArrayList stylesClass 
                case "style":
                    InitClassStyle(child);
                    break;
                //JPV : gestion text
                case "text":
                    converted = ConvertText(elementStyle as SvgTextElement);
                    break;
                case "rect":
                    converted = ConvertRect(elementStyle as SvgRectElement);
					break;
				case "ellipse":
                    converted = ConvertEllipse(elementStyle as SvgEllipseElement);
					break;
				case "g":
                    converted = ConvertGroup(elementStyle as SvgGroupElement);
					break;
				case "line":
                    converted = ConvertLine(elementStyle as SvgLineElement);
					break;
				case "polyline":
                    converted = ConvertPolyline(elementStyle as SvgPolylineElement);
					break;
				case "path":
                    converted = ConvertPath(elementStyle as SvgPathElement);
					break;
				case "polygon":
                    converted = ConvertPolygon(elementStyle as SvgPolygonElement);
					break;                    
			}
			if (converted != null)
			{
				SvgStyledTransformedElement styled = child as SvgStyledTransformedElement;
				if (styled != null && styled.Transform != null)
					converted.TransformBy(ConvertTransformList(styled.Transform));
				string name = child.Id;
				if (name != null && name.Length > 0)
				{
					// names that begin with numbers don't work well in Visual Studio designers
					if (char.IsNumber(name[0]))
						name = "id" + name;
					// other characters don't work well either
					if (name.Contains("-"))
						name = name.Replace('-', '_');
					converted.Name = name;
				}
			}
			return converted;
		}

        /// <summary>
        /// JPV : Create a PointText from a text element with TSpan
        /// </summary>
        /// <param name="svg">Element to trait</param>
        /// <returns>PointText created</returns>
        private Prodige.Drawing.PointText ConvertText(SvgTextElement svg)
        {
            Prodige.Drawing.PointText pt = new Prodige.Drawing.PointText();
            pt.Text = svg.Text;
            pt.Location = new Vector(svg.X, svg.Y);
            ConvertStrokeFill(svg, pt);

            pt.TextAppearance = ConvertFontStyle(svg);

            return pt;
        }

        private Prodige.Drawing.Rectangle ConvertRect(SvgRectElement svg)
        {
            Prodige.Drawing.Rectangle shape = new Prodige.Drawing.Rectangle();
            shape.Bounds = new FRectangle(svg.X, svg.Y, svg.Width, svg.Height);
            shape.CornerRadii = new Radii(svg.RX, svg.RY);
            ConvertStrokeFill(svg, shape);
            return shape;
        }

        private Ellipse ConvertEllipse(SvgEllipseElement svg)
        {
            Ellipse shape = new Ellipse();
            shape.Bounds = new FRectangle(svg.CX.Value - svg.RX.Value, 
				svg.CY.Value - svg.RY.Value, svg.RX.Value * 2, svg.RY.Value * 2);
            ConvertStrokeFill(svg, shape);
            return shape;
        }

        private Polyline ConvertLine(SvgLineElement svg)
        {
            Polyline shape = new Polyline();
            shape.Points.Add(new Vector(svg.X1.Value, svg.Y1.Value));
            shape.Points.Add(new Vector(svg.X2.Value, svg.Y1.Value));
            ConvertStrokeFill(svg, shape);
            return shape;
        }

        private Polyline ConvertPolyline(SvgPolylineElement svg)
        {
            // NOTE: Polylines are "open" ^.^
            Polyline shape = new Polyline();
            SvgPoints points = svg.Points;
            foreach (PointF pt in points._pts)
                shape.Points.Add(pt);
            ConvertStrokeFill(svg, shape);
            return shape;
        }

        private Polyline ConvertPolygon(SvgPolygonElement svg)
        {
            Polyline shape = new Polyline();
            SvgPoints points = svg.Points;
            foreach (PointF pt in points._pts)
                shape.Points.Add(pt);
            ConvertStrokeFill(svg, shape);
            //Always closed, Polygons are.
			shape.Closed = true; // ??
            return shape;
        }

		/// <summary>Current point, when converting path data.</summary>
		private Vector current;
		/// <summary>Indicates coordinates are relative, when converting path data.</summary>
		private bool relative;
		/// <summary>Current path segment.</summary>
		private PathSeg seg;

		private Path ConvertPath(SvgPathElement svg)
		{            
			Path shape = new Path();
			PathPointCollection points = shape.PathPoints;
			float x, y;
			current = Vector.Empty;
			SvgPath d = svg.D;
			for (int i = 0; i < d.Count; i++)
			{
				seg = d[i];
				relative = char.IsLower(seg.Char[0]);
				switch (seg.Type)
				{
					case SvgPathSegType.SVG_SEGTYPE_MOVETO:
						for (int j = 0; j * 2 < seg.Data.Length; j++)
							points.Add(new PathPoint(GetPoint(j), PointType.Start));
						break;
					case SvgPathSegType.SVG_SEGTYPE_CLOSEPATH:
						//points.Add(new PathPoint().Type = PointType.EndLine);
						break;
					case SvgPathSegType.SVG_SEGTYPE_LINETO:
						for (int j = 0; j * 2 < seg.Data.Length; j++)
							points.Add(new PathPoint(GetPoint(j), PointType.EndLine));
						break;
					case SvgPathSegType.SVG_SEGTYPE_CURVETO:
						for (int j = 0; j * 2 < seg.Data.Length; j += 3)
						{
							points.Add(new PathPoint(GetPoint(j), PointType.Control1));
							points.Add(new PathPoint(GetPoint(j + 1), PointType.Control2));
							points.Add(new PathPoint(GetPoint(j + 2), PointType.EndBezier));
						}
						break;
					case SvgPathSegType.SVG_SEGTYPE_SMOOTHCURVETO:
						Segment last = points.Segments.Last;
						Vector control;
						if (last.Type == SegmentType.Bezier)
							control = last.Point - (last.Control2 - last.Point);
						else
							control = last.Point;
						for (int j = 0; j * 2 < seg.Data.Length; j += 2)
						{
							points.Add(new PathPoint(control, PointType.Control1));
							Vector control2 = GetPoint(j);
							points.Add(new PathPoint(control2, PointType.Control2));
							Vector end = GetPoint(j + 1);
							points.Add(new PathPoint(end, PointType.EndBezier));
							control = end - (control2 - end);
						}
						break;
					case SvgPathSegType.SVG_SEGTYPE_VLINETO:
						x = current.X;
						for (int j = 0; j < seg.Data.Length; ++j)
						{
							y = seg.Data[j];
							if (relative)
								y += current.Y;
							points.Add(new PathPoint(x, y, PointType.EndLine));
						}
						break;
					case SvgPathSegType.SVG_SEGTYPE_HLINETO:
						y = current.Y;
						for (int j = 0; j < seg.Data.Length; ++j)
						{
							x = seg.Data[j];
							if (relative)
								x += current.X;
							points.Add(new PathPoint(x, y, PointType.EndLine));
						}
						break;
				}
				current = points.Last.Point;
			}
			ConvertStrokeFill(svg, shape);
			return shape;
		}

		/// <summary>
		/// Take the two relative coordinates from the array of floats
		/// at the specified index, and return an absolute point
		/// </summary>
		/// <param name="index">point index -- index in array will be twice this</param>
		/// <returns>an absolute point</returns>
		private Vector GetPoint(int pointIndex)
		{
			float[] data = seg.Data;
			int i = pointIndex * 2;
			Vector v = new Vector(data[i], data[i + 1]);
			if (relative)
				return current + v;
			return v;
		}

        /// <summary>
        /// JPV : Apply a text appearance from a font style
        /// </summary>
        /// <param name="svg">Element with the style to format</param>
        /// <returns>Text Style modified</returns>
        private TextAppearance ConvertFontStyle(SvgTextElement svg)
        {
            TextAppearance textStyle = new TextAppearance();
            
            if (svg != null)
            {
                textStyle = ConvertTextAppearanceStyle(textStyle, svg.Style);
                textStyle = ConvertTextAppearanceStyle(textStyle, svg.StyleTSpan);
            }
                       
            return textStyle;
        }

        /// <summary>
        /// JPV : Apply a text appearance from a font style
        /// </summary>
        /// <param name="textStyle">Text Style to modify</param>
        /// <param name="svg">Element with the style to format</param>
        /// <returns>Text Style modified</returns>
        private TextAppearance ConvertTextAppearanceStyle(TextAppearance textStyle, SvgStyle svg)
        {
            if (textStyle == null)
                textStyle = new TextAppearance();

            if (svg != null)
            {
                string style = (string)svg["font-style"];
                if (style != null)
                {
                    if (style.Equals("italic"))
                    {
                        textStyle.Italic = true;
                    }
                }

                string decoration = (string)svg["text-decoration"];
                if (decoration != null)
                {
                    if (decoration.Equals("underline"))
                    {
                        textStyle.Underline = true;
                    }
                }

                string weight = (string)svg["font-weight"];
                if (weight != null)
                {
                    if (weight.Equals("bold"))
                    {
                        textStyle.Bold = true;
                    }
                }

                string size = (string)svg["font-size"];
                if (size != null)
                {
                    if (size.IndexOf("px") > -1)
                    {
                        textStyle.SizeUnit = GraphicsUnit.Pixel;
                        textStyle.Size = float.Parse(size.Substring(0, size.IndexOf("px")));
                    }
                }

                string name = (string)svg["font-family"];
                if (name != null)
                {
                    textStyle.FaceName = name;
                }
            }

            return textStyle;
        }
        
        private void ConvertStrokeFill(SvgStyledTransformedElement svg, Element element)
        {
            Style style = new Style();
            style.Name = svg.Id;
            style.Stroke = new Stroke();
			bool hasFill = false;
			bool hasStroke = false;
			Fill fill = null;
            
            object svgFill = GetStyleAttribute(svg, "fill");
			if (svgFill != null && svgFill as string != "none")
			{
				SvgColor color = svgFill as SvgColor;
				if (color != null)
					fill = new SolidFill(color.Color);
				SvgLinearGradient lg = svgFill as SvgLinearGradient;
				if (lg != null)
					fill = ConvertLinearGradient(lg, element);
				object opacity = GetStyleAttribute(svg, "fill-opacity");
				if (opacity != null)
				{
					if (fill == null)
						fill = new SolidFill();
					fill.Opacity = (float)opacity;
				}
				if (fill != null)
				{
					style.Fill = fill;
					hasFill = true;
				}
			}

            object stroke = GetStyleAttribute(svg, "stroke");
			if (stroke != null && stroke as string != "none")
			{
				SvgColor strokeColor = stroke as SvgColor;
				if (strokeColor != null)
				{
					style.Stroke = new Stroke(strokeColor.Color);
					hasStroke = true;
				}
				object strokeOpacity = GetStyleAttribute(svg, "stroke-opacity");
				if (strokeOpacity != null)
					style.Stroke.Opacity = (float)strokeOpacity;
				SvgLength width = GetStyleAttribute(svg, "stroke-width") as SvgLength;
				if (width != null)
					style.Stroke.Width = width.Value;
			}

			if (hasStroke || hasFill)
			{
				element.Style = style.Name;
				styles.Add(style);
			}
			Shape shape = element as Shape;
			if (shape != null)
			{
				if (hasStroke && hasFill)
					shape.DrawAction = ShapeDrawAction.EdgeAndFill;
				else if (hasStroke)
					shape.DrawAction = ShapeDrawAction.Edge;
				else if (hasFill)
					shape.DrawAction = ShapeDrawAction.Fill;
				else
					shape.Visible = false;
			}
        }

		private static LinearGradientFill ConvertLinearGradient(SvgLinearGradient lg,
			Element element)
		{
			Vector p1 = new Vector(lg.X1.Value, lg.Y1.Value);
			Vector p2 = new Vector(lg.X2.Value, lg.Y2.Value);
			// get as FMatrix, transform points
			FMatrix m = ConvertTransformList(lg.GradientTransform);
			if (!m.IsIdentity)
			{
				p1 = m.TransformPoint(p1);
				p2 = m.TransformPoint(p2);
			}
			if (lg.GradientUnits == "userSpaceOnUse")
			{
				// put these points into relative coordinates
				FRectangle bounds = element.Bounds;
				p1 = p1.AbsoluteToRelativePoint(bounds);
				p2 = p2.AbsoluteToRelativePoint(bounds);
			}
			LinearGradientFill lgf = new LinearGradientFill(p1, p2, 
				FillCoordinateType.Relative, GradientType.MultipleColor);
			// TODO: how is gamma correction set in svg?
			lgf.GammaCorrection = false;
			lgf.GradientColors.Clear();
			foreach (SvgElement child in lg.Children)
			{
				SvgStopElement stop = child as SvgStopElement;
				if (stop == null)
					continue;
				SvgColor svgColor = GetStyleAttribute(stop, "stop-color") as SvgColor;
				if (svgColor != null)
				{
					float fo = 1;
					object stopOpacity = GetStyleAttribute(stop, "stop-opacity");
					if (stopOpacity != null)
						fo = (float)stopOpacity;
					lgf.GradientColors.Add(new GradientColor(stop.Offset, svgColor.Color, fo));
				}
			}
			if (lgf.GradientColors.Count > 0)
			{
				// go through and make sure that GradientColor Position 0 and 1 are both set
				// if not, fix. Otherwise GDI+ will throw and exception during display.
				GradientColorCollection gcc = lgf.GradientColors;
				bool zero = false;
				bool one = false;
				GradientColor smallest = new GradientColor(1, Color.White);
				GradientColor largest = new GradientColor(0, Color.Black);
				foreach (GradientColor gc in gcc)
				{
					if (gc.Position == 0)
						zero = true;
					else if (gc.Position == 1)
						one = true;
					if (gc.Position < smallest.Position)
						smallest = gc;
					else if (gc.Position > largest.Position)
						largest = gc;
				}
				if (!zero)
					gcc.Add(new GradientColor(0, smallest.Color, smallest.Opacity));
				if (!one)
					gcc.Add(new GradientColor(1, largest.Color, largest.Opacity));
			}	
			return lgf;
		}

		private static object GetStyleAttribute(SvgStyledTransformedElement svg, string key)
		{
			object value = svg[key];
			if (value == null)
				value = svg.Style.Get(key);
			if (value == null)
			{	
				SvgStyledTransformedElement styledParent = svg.Parent as SvgStyledTransformedElement;
				if (styledParent != null)
					return GetStyleAttribute(styledParent, key);
			}
			return value;
		}

		private static FMatrix ConvertTransformList(SvgTransformList transformList)
		{
			FMatrix m = FMatrix.Identity;
			if (transformList == null)
				return m;
			for (int i = 0; i < transformList.Count; ++i)
				m = new FMatrix(transformList[i].Matrix) * m;
			return m;
		}
    }
}
