/*
	Copyright c 2003 by RiskCare Ltd.  All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
	LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
	OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/


using System;
using System.Collections;
using System.IO;
using System.Xml;
using System.Reflection;
using SvgNet.SvgElements;
using SvgNet.SvgTypes;
using System.Text;

namespace SvgNet
{
	/// <summary>
	/// Static methods to produce/write/copy Svg documents reside in this class.
	/// </summary>
	public class SvgFactory
	{
		private SvgFactory() {}

        //JPV : true if it is a style element for CSS description with CDATA
        private static bool inStyle = false;

		private static Hashtable _elementNameDictionary;
		private static readonly object[] _emptyArgs = new object[0];
		private static readonly Type[] _emptyTypes = new Type[0];

		/// <summary>
		/// Given an xml document and (optionally) a particular element to start from, read the xml nodes
		/// and construct a tree of <see cref="SvgElement"/> objects.  Xml tags that do not correspond
		/// to a particular class will be represented by an <see cref="SvgGenericElement"/>.  This means
		/// any XML input can be read in and written out, even if it has nothing to do with Svg.
		/// More usefully, it means that new and unsupported tags and attributes will be represented
		/// in the <c>SvgElement</c> tree and written out correctly even if SvgNet does
		/// not understand them.
		/// </summary>
		/// <param name="doc"></param>
		/// <param name="el"></param>
		/// <returns></returns>
		public static SvgElement LoadFromXML(XmlDocument doc, XmlElement el)
		{
			if (el == null)
			{
				foreach(XmlNode noddo in doc.ChildNodes)
				{
					if (noddo.GetType() == typeof(XmlElement))
					{
						el = (XmlElement)noddo;
						break;
					}
				}
			}
			if (el == null)
				return null;
			// NOTE: this code uses a dictionary of types to determine which object to create.
			// A less clumsy approach that does not require reflection would be to create a dictionary
			// of prototypes that are cloned when a new object is needed.
			if (_elementNameDictionary == null)
				BuildElementNameDictionary();
			Type t = (Type)_elementNameDictionary[el.Name];
			SvgElement e = (SvgElement)t.GetConstructor(_emptyTypes).Invoke(_emptyArgs);
			Hashtable ids = new Hashtable();
			RecLoadFromXML(e, doc, el, ids);
			ResolveUris(e, ids);
			return e;
		}

		/// <summary>
		/// Used by LoadFromXML
		/// </summary>
		/// <param name="e"></param>
		/// <param name="doc"></param>
		/// <param name="el"></param>
		private static void RecLoadFromXML(SvgElement e, XmlDocument doc, XmlElement el, Hashtable ids)
		{
			e.ReadXmlElement(doc, el);
			string id = e.Id;
			if (id != null && id != "")
				ids[id] = e;
			foreach(XmlNode noddo in el.ChildNodes)
			{
				if (noddo.GetType() == typeof(XmlElement))
				{
					XmlElement childXml = (XmlElement)noddo;
					Type t = (Type)_elementNameDictionary[childXml.Name];
					SvgElement childSvg;
					if (t == null)
					{
					    //JPV : Test if it is a style element
					    if (noddo.Name.ToUpper() == "STYLE")
					        inStyle = true;
						childSvg = new SvgGenericElement(childXml.Name);
					}
					else
						childSvg = (SvgElement)t.GetConstructor(_emptyTypes).Invoke(_emptyArgs);
					e.AddChild(childSvg);
					RecLoadFromXML(childSvg, doc, childXml, ids);
				}
				else if (noddo.GetType() == typeof(XmlText))
				{
					XmlText xt = (XmlText)noddo;
					TextNode tn = new TextNode(xt.InnerText);
					e.AddChild(tn);
				}
                //JPV : Get the CDATA text to define a class of CSS style
                else if (noddo.GetType() == typeof(XmlCDataSection) && inStyle)
                {
                    string cssStyle = noddo.Value.Trim();

                    char[] sep = new char[2];
                    sep[0] = '{';
                    sep[1] = '}';

                    string[] styles = cssStyle.Split(sep, StringSplitOptions.RemoveEmptyEntries);

                    for (int i = 0; i < styles.Length; i = i + 2)
                    {
                        string name = styles[i].Substring(styles[i].IndexOf('.')).TrimEnd();
                        string sstyle = styles[i + 1];
                        SvgClassStyleElement classStyle = new SvgClassStyleElement();
                        classStyle.ClassName = name;
                        classStyle.Style = new SvgStyle(sstyle);
                        e.AddChild(classStyle);
                    }
                    inStyle = false;
                }
                //End JPV
            }
		}

		/// <summary>A temporary list used in ResolveUris.</summary>
		private static readonly ArrayList _replace = new ArrayList();

		/// <summary>Represents an attribute that needs to be replace in ResolveUris.</summary>
		private sealed class AttributeReplacement
		{
			public readonly object Key;
			public readonly object Value;
			public AttributeReplacement(object key, object value)
			{
				Key = key;
				Value = value;
			}
		}

		/// <summary>
		/// Find all "url(#...)" strings in attributes and resolve them to object references.
		/// </summary>
		/// <param name="e"></param>
		/// <param name="ids"></param>
		private static void ResolveUris(SvgElement e, Hashtable ids)
		{
			string pattern = "url(#";
			int patternLength = pattern.Length;
			// find all attribute values that need to be replaced
			foreach (DictionaryEntry entry in e.Attributes)
			{
				string uri = entry.Value as string;
				if (uri != null && uri.StartsWith(pattern) && uri.Length > patternLength + 2)
				{
					// try to find a match in ids
					uri = uri.Substring(patternLength, uri.Length - patternLength - 1);
					_replace.Add(new AttributeReplacement(entry.Key, ids[uri]));
				}
			}
			// replace
			foreach (AttributeReplacement ar in _replace)
				e[ar.Key as string] = ar.Value;
			_replace.Clear();
			foreach (SvgElement child in e.Children)
				ResolveUris(child, ids);
		}

		/// <summary>
		/// Used by LoadFromXML
		/// </summary>
		private static void BuildElementNameDictionary()
		{

			_elementNameDictionary = new Hashtable();
			Assembly asm = Assembly.GetExecutingAssembly();
			Type[] ta = asm.GetExportedTypes();
			foreach(Type t in ta)
			{
				if (t.IsSubclassOf(typeof(SvgElement)))
				{
					SvgElement e = (SvgElement)t.GetConstructor(new System.Type[0]).Invoke(new object[0]);
					_elementNameDictionary[e.Name] = e.GetType();
				}
			}
		}

		/// <summary>
		/// Helper function to compress long xml attributes into entities.
		/// <para>
		/// This would work on any XML, it is not SVG specific, so it should eventually be in some 'xml tools' class.
		/// </para>
		/// </summary>
		/// <param name="doc"></param>
		/// <param name="el"></param>
		/// <returns>A string of entities which can be inserted into the DOCTYPE when the document is written.</returns>
		public static string CompressXML(XmlDocument doc, XmlElement el)
		{

			Hashtable entities = new Hashtable();
			Hashtable counts = new Hashtable();

			int idx = 0;
			RecCompXML(counts, entities, doc, el, ref idx);

			string ents = "";
			StringBuilder sbEnts = new StringBuilder();

			foreach(string key in entities.Keys)
			{
				sbEnts.Append("\r\n\t<!ENTITY ");
				sbEnts.Append(entities[key]);
				sbEnts.Append(" '");
				sbEnts.Append(key.Replace("%", "&#37;")); //I guess we need to escape fully, but surely no other special char can appear?)
				sbEnts.Append("'>");
			}

			if (sbEnts.Length>0) sbEnts.Append("\r\n");
			ents = sbEnts.ToString();

			return ents;
		}

		/// <summary>
		/// Used by CompressXML
		/// </summary>
		/// <param name="counts">Map of attribute to number of occurrences -- could be used to improve algorithm</param>
		/// <param name="entities">Map of attribute to entity name</param>
		/// <param name="doc"></param>
		/// <param name="el"></param>
		/// <param name="idx">Number that is incremented to provide new entity names</param>
		private static void RecCompXML(Hashtable counts, Hashtable entities, XmlDocument doc, XmlElement el, ref int idx)
		{	
			ArrayList keys = new ArrayList();

			foreach(XmlAttribute att in el.Attributes)
			{
				//don't try to do anything with xmlns nodes.
				if (att.Name.IndexOf("xmlns") == -1)
					keys.Add(att.Name);
			}

			foreach(string s in keys)
			{
				string val = el.Attributes[s].Value;
	
				if (counts[val] == null)
					counts[val] = 1;
				else
					counts[val] = (int)counts[val] + 1;

				if (val.Length > 30)
				{
					string entname;

					if(entities[val] == null)
					{
						idx += 1;
						entname = "E"+idx.ToString();
						entities[val] = entname;
					}
					else
						entname = (string)entities[val];

					XmlAttribute attr;
					
					//xlinks are a special case at the moment as the .NET XML API requires whatever bit of code
					//outputs a qualified name to *just magically happen to know* the URI that the prefix refers to.
					if (s.IndexOf("xlink") == 0)
						attr = doc.CreateAttribute(s, "http://www.w3.org/1999/xlink");
					else
						attr = doc.CreateAttribute(s);

					attr.AppendChild(doc.CreateEntityReference(entname));
					el.SetAttributeNode(attr);
				}
			}

			foreach(XmlNode ch in el.ChildNodes)
			{
				if (ch.GetType() == typeof(XmlElement))
					RecCompXML(counts, entities, doc, (XmlElement)ch, ref idx);
			}
		}
	
		/// <summary>
		/// Create a complete deep copy of the given tree of <c>SvgElement</c> objects.
		/// A new set of elements is created, and if the attributes are cloneable they are deep-copied too.
		/// Since strings and all SvgType classes are cloneable, the new tree is independant of the old.
		/// </summary>
		/// <param name="el"></param>
		/// <returns></returns>
		public static SvgElement CloneElement(SvgElement el)
		{
			SvgElement clone = (SvgElement)el.GetType().GetConstructor(new System.Type[0]).Invoke(new object[0]);
			foreach(string key in el.Attributes.Keys)
			{
				object o = el[key];
				if (typeof(ICloneable).IsInstanceOfType(o))
					clone[key] = ((ICloneable)o).Clone();
				else
					clone[key] = o;
			}
			foreach(SvgElement ch in el.Children)
				clone.AddChild(CloneElement(ch));
			return clone;
		}
	}
}
