/*
	Portions Copyright c 2003 by RiskCare Ltd.  All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
	OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
	LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
	OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/
// Copyright c 2005 Prodige Software Corporation.  All rights reserved.
// Redistribution not permitted without permission.


using System;
using System.Collections;
using System.Xml;
using System.Text;
using System.IO;
using SvgNet.SvgTypes;

namespace SvgNet
{
	/// <summary>
	/// This is an SvgElement that can have a CSS style and an SVG transformation list.  It contains
	/// special properties to make reading and setting the style and the transformation easier.  All 
	/// SVG elements that actually represent visual entities or groups of entities are 
	/// <c>SvgStyledTransformedElements</c>.
	/// </summary>
	public class SvgStyledTransformedElement : SvgElement
	{
		public SvgStyledTransformedElement() {}

		public SvgStyledTransformedElement(string id) : base(id) {}

		protected override object ConvertAttribute(string attributeName, string stringValue)
		{
			if (attributeName == "style")
				return new SvgStyle(stringValue);	
			if (attributeName == "transform")
				return new SvgTransformList(stringValue);
			object value = SvgStyle.ConvertAttribute(attributeName, stringValue);
			if (!(value is string))
				return value;
			return base.ConvertAttribute(attributeName, stringValue);
		}

		/// <summary>
		/// Overridden in this class to provide special handling for the style and transform attributes,
		/// which are often long and complicated.  For instance, it may be desirable for styles to be 
		/// written as entities or as separate attributes.
		/// </summary>
		/// <param name="doc"></param>
		/// <param name="parent"></param>
		public override void WriteXmlElements(XmlDocument doc, XmlElement parent)
		{
			XmlElement me = doc.CreateElement("", Name, doc.NamespaceURI);

			foreach(string s in _atts.Keys)
			{
				
				if (s == "style")
					WriteStyle(doc, me, _atts[s]);
				else if (s == "transform")
					WriteTransform(doc, me, _atts[s]);
				else
				{
					//xlink nodes are a special case because we have to know the xlink URI.
					//eventually there will have to be some system for identifying qualified names and looking up the URI.
					//why it is not possible to simply insert 'xlink:href' as an attribute name in this API is utterly beyond me.
					if (s.IndexOf("xlink")==0)
					{
						XmlAttribute att = (XmlAttribute)doc.CreateAttribute(s, "http://www.w3.org/1999/xlink");
						att.Value = _atts[s].ToString();
						me.SetAttributeNode(att);
					}
					else
						me.SetAttribute(s, doc.NamespaceURI, _atts[s].ToString());
				}
			}
			foreach(SvgElement el in _children)
				el.WriteXmlElements(doc, me);

			if(parent == null)
				doc.AppendChild(me);
			else
				parent.AppendChild(me);
		}

		/// <summary>
		/// Provides an easy way to get the attribute called "style" as an <c>SvgStyle</c> object.  
		/// If no such attribute has been set, one is created when this property is read.
		/// </summary>
		public virtual SvgStyle Style
		{
			get
			{
				SvgStyle st = _atts["style"] as SvgStyle;
				if (st == null)
				{
					st = new SvgStyle();
					_atts["style"] = st;
				}
				return st;
			}
			set { _atts["style"] = value; }
		}

		/// <summary>
		/// Provides an easy way to get the attribute called "transform" as an 
		/// <c>SvgTransformList</c> object.  If no such attribute has been set, one is created when
		/// this property is read.
		/// </summary>
		public SvgTransformList Transform
		{
			get
			{
				SvgTransformList tl = _atts["transform"] as SvgTransformList;
				if (tl == null)
				{
					tl = new SvgTransformList();
					_atts["transform"] = tl;
				}
				return tl;
			}
			set{ _atts["transform"] = value; }
		}

		private void WriteStyle(XmlDocument doc, XmlElement me, object o)
		{
			if (o.GetType() != typeof(SvgStyle))
			{
				me.SetAttribute("style", doc.NamespaceURI, o.ToString());
				return;
			}

			SvgStyle style = (SvgStyle)o;

			/*
			foreach(string s in style.Keys)
			{
				me.SetAttribute(s, doc.NamespaceURI, style.Get(s).ToString());
			}
			*/

			me.SetAttribute("style", doc.NamespaceURI, style.ToString());

			doc.CreateEntityReference("pingu");
		}

		private void WriteTransform(XmlDocument doc, XmlElement me, object o)
		{
			//if (o.GetType() != typeof(SvgTransformList))
			//{
				me.SetAttribute("transform", doc.NamespaceURI, o.ToString());
				return;
			//}

		}
	}
}
